/**
 *  file     TraderImplementation.java
 *  package
 *  platform Java 1.4.1 ; CS Tux / Solaris
 *
 *  project  CS537, Fall 2002, Project2
 *  author   Wong, Kin-Chung (kin-chun) (ryan90s@hotmail.com)
 *  team     (data, kin-chun)
 *  
 *  date_created            2002 Sep 24
 *  date_last_modified      2002 Oct 05 : not finalized
 *
 *  version model #1
 *  version detail: (about buy method)
 *    Sequentially request each commodity as follows: 
 *    (assume specialty is coal. Order of Operation is fixed for all trader.)
 *
 *    Reserve enough coal as requested.
 *    Reserve any oil. Secure enough coal and trade. Reserve traded oil.
 *    Reserve any elec. Secure enough coal and trade. Reserve traded elec.
 *    Reserve any gas. Secure enough coal and trade. Reserve traded gas.
 *    Deliver all 4 commodity.
 */


/**
 *  An instantiable implementation of the Trader class. It is a monitor
 *  class, that is, it does not act on its own. It carries out operations
 *  only when requested from customers and suppliers. All methods in
 *  this class are thread-safe.
 */
public class TraderImplementation implements Enron, Trader {

    // Short names of commodity. (used for message printing only)
    private String cname_s[]={"coal", "oil", "gas", "elec"};

    // The commodity that this trader is specialized in.
    private int specialty;

    // Physical amount of stock in hand.
    private int[] stock;

    // The amount of stock that is available 
    // (ie. not reserved by another customer)
    // Note: reservation is made only after the stock really exists.
    // otherwise the customer waits before making reservation.
    private int[] surplus;

    /** Constructor. 
     *  The Trader will not own any commodities initially.
     *  
     *  @param specialty The code for the commodity which this Trader 
     *                   is specialized in.
     */
    public TraderImplementation (int specialty) {
	stock= new int[COMMODITIES];
	surplus= new int[COMMODITIES];
	for (int i=0; i<stock.length; i++) {
	    stock[i]=0;
	    surplus[i]=0;
	}
	this.specialty = specialty;
    }

    /** 
     *  Requests commodities from the Trader. This method is to be called
     *  by a Customer.
     *
     *  @param amount The amount of each commodity to be purchased.
     *                Array of size (Enron.COMMODITIES), supplied by caller.
     */
    public void buy(int [] amount) {
	Project2.debug("entering buy");
	// For each good, allocate the amount requested
	// but don't subtract them from record.
	// Stocks that are "reserved" are subtracted from  
	// the surplus variable. There may be delays but it's thread safe.
	//
	for (int item=0; item<COMMODITIES; item++) {
	    
	    // we do not need to handle zero request.
	    if (amount[item]==0) continue;
	    Project2.debug(cname[specialty]+".buy(): trying to allocate "+amount[item]+" of "+cname[item]);
	    printCurrentBalance();
	    if (item==specialty) {
		// Processing specialty request.
		// wait until there's enough of it
		// Reserve that amount.
		reserve(item, amount[item]);		
		
	    } else {
		// Processing non-specialty request.
		// Wait until there's either enough non-specialty or 
		// specialty commodity.
		//
		// note: if non-specialty is ready it is notified by trade() 
		//       if specialty is ready it is notified by supply()
		int remain = amount[item] - reserveUpTo(item, amount[item]);
		reserve(specialty, remain);  // reserve for trade
		tradeWith(item, remain);  // trade
		// the newly available amount is already reserved
	    } // end allocate_Non_Specialty
	    
	} // end for (item) loop
	
	// By now, all required goods are ready and it's time to return.
	printCurrentBalance();
	for (int item=0; item<COMMODITIES; item++) {
	    deliver(item, amount[item]);
	}
	Project2.debug("Customer is satisfied. Leaving buy");
	printCurrentBalance();
    }
    
    /** void tradeWith(int, int)
     *  Trades with another trader and atomically update the local stock
     *  when returned from call.
     *  Always called from a synchronized local method, namely, "buy".
     */
    private void tradeWith(int commodity, int amount) {
	Project2.debug("entering tradeWith");
	if (amount == 0) return;
	if (commodity == specialty) return;
	Project2.specialist(commodity).trade(this.specialty, amount);
	deliver(this.specialty, amount);
	receiveAsReserved(commodity, amount);
	Project2.debug("leaving tradeWith");
    }
    
    
    /**
     *  Accepts trade requests from other traders.
     *  Always called from a synchronized outside method, namely "tradeWith"
     */
    public void trade(int what, int amount) {
	Project2.debug("entering trade");
	if (amount == 0) return;
	reserve(this.specialty, amount);
	deliver(this.specialty, amount);
	receive(what, amount);
	Project2.debug("leaving trade");
    }
    
    /**
     *  Deposits some amount of the commodity, of which this Trader
     *  has specialty in.
     *
     *  @param amount The amount to deposit. The commodity type is 
     *                the same as the one this Trader has specialty in.
     */
    public void supply(int amount) {
	receive(this.specialty, amount);
    }
    
    /**
     *  Subtract (reserve) the amount from the surplus. Reserved stocks
     *  are no longer usable by any other customers / traders except
     *  the current Customer (the caller of this thread). Synchronized.
     *
     *  If this amount is not available, it will wait until
     *  it is available.
     *
     *  @param item The type of commodity to be reserved
     *  @param amount The amount to be reserved
     */
    private synchronized void reserve(int item, int amount) {
	while (surplus[item]<amount) {
	    try {
		wait();
	    } catch (InterruptedException ex) {
		Thread.currentThread().stop();
	    }
	}
	surplus[item] -= amount;
    }


    /**
     *  Subtract (reserve) as much as it can from the surplus, and
     *  report the amount that is actually secured.
     *  Synchronized. This method does not wait.
     *
     *  If this amount is not available, it will report zero (0).
     *
     *  @param item The type of commodity to be reserved
     *  @param amount The amount to be reserved
     */
    private synchronized int reserveUpTo(int item, int amount) {
	int amountReserved=0;
	if (amount <= surplus[item]) {
	    // enough to satisfy all
	    amountReserved = amount;
	} else {
	    // not enough -> attempt to secure whatever is available.
	    amountReserved = surplus[item];
	}
	surplus[item] -= amountReserved;
	return amountReserved;
    }

    /**
     *  Increase the stock amount as specified. Synchronized.
     *  
     *  @param item The type of specialty received
     *  @param amount The amount received
     */
    private synchronized void receive(int item, int amount) {
	Project2.debug(cname[specialty]+" trader received "+amount
		       +" of "+cname_s[item]);
	this.stock[item] += amount;
	this.surplus[item] += amount;
	printCurrentBalance();
	notifyAll();
    }

    /**
     *  Increase the stock amount as specified. Automatically reserve
     *  this amount so that only the current customer can use it.
     *  Synchronized.
     *  
     *  @param item The type of specialty received
     *  @param amount The amount received
     */
    private synchronized void receiveAsReserved(int item, int amount) {
	Project2.debug(cname_s[specialty]+" trader received "+amount
		       +" of "+cname_s[item]);
	this.stock[item] += amount;
    }

    /**
     *  Decrease the stock amount as specified. Synchronized.
     *  Before calling this method, the amount has to be subtracted from
     *  the "surplus" variable first.
     *  
     *  @param item The type of specialty taken away
     *  @param amount The amount taken away
     */
    private synchronized void deliver(int item, int amount) {
	Project2.debug(cname_s[specialty]+" trader delivered "+amount
		       +" of "+cname_s[item]);
	this.stock[item] -= amount;
    }

    /**
     *  Returns the physical amount of stock in hand. Used by debugger.
     *
     *  @param result An array of size Enron.COMMODITIES (declared by caller).
     *                It will contain the stock amount when returned from
     *                the method.
     */    
    public synchronized void getAmountOnHand(int[] result) {
	for (int i=0; i<COMMODITIES; i++) {
	    result[i]=stock[i];
	}
	Project2.debug(cname[specialty]+" trader\'s surplus variable:");
	for (int item=0; item<COMMODITIES; item++) {
	    Project2.debug("\t"+surplus[item]+" of "+cname[item]);
	}
    }

    /**
     *  Prints out the amount of stock to the debug stream.
     */
    private synchronized void printCurrentBalance() {
	StringBuffer s=new StringBuffer();
	s.append("(");
	s.append(cname_s[specialty]);
	s.append(") ");
	for (int item=0; item<COMMODITIES; item++) {
	    s.append(cname_s[item]);
	    s.append(":"+stock[item]);
	    s.append("/"+surplus[item]);
	    s.append("/"+0);
	    s.append("/"+(stock[item]-surplus[item]));
	    if (item<COMMODITIES-1) 
		s.append(" ");
	}
	Project2.debug(s.toString());
    }
}



