/** CS 537, Fall 2002, Project 3.
 * An order for commodities.  It keeps track of the amounts requested and
 * serves as a waiting place for the ordering thread to wait for them.
 * <p>Copyright &#169; 2002 by Marvin Solomon.  All rights reserved.</p>
 * @version $Revision: 1.6 $ $Date: 2002/10/17 10:50:05 $
 */
public class Order {
    /** Source of sequence numbers. */
    private static int nextSeq = 0;

    /** The original amount requested. */
    private int[] request;

    /** The requesting customer's id (for debugging output). */
    private int id;

    /** The sequence number of this request in the order received. */
    public final int seq;

    /** The amount allocated thus far. */
    private int[] alloc;

    /** Indication that this request has been released by calling release(). */
    private boolean done = false;

    /** Return code to be returned by await.  The value false indicates
     * the request was refused (because the system is shutting down).
     */
    private boolean result;

    /** Creates a new Order.
     * Note:  This method is not thread-safe, so it should be only called from
     * a synchronized method that prevents two concurrent invocations of
     * "new Order(...)".
     * @param id the requesting customer's id (for debugging output).
     * @param request the amount of each commodity needed.
     */
    public Order(int id, int[] request) {
        this.id = id;
        this.request = (int[]) request.clone();
        this.seq = ++nextSeq;
        alloc = new int[request.length];
    } // Order(int[])

    /** Check whether the remainder of this request is less than or equal to
     * amt (in all commodities).
     * @param amt the amount to compare to.
     * @return true if the request can be completely satisfied by amt.
     */
    public synchronized boolean lessOrEqual(int[] amt) {
        for (int i = 0; i < request.length; i++) {
            if (request[i] - alloc[i] > amt[i]) {
                return false;
            }
        }
        return true;
    } // lessOrEqual(int[])

    /** Give some resources to this Order from supply.
     * The amount of each resource given is the minimum of the amount in supply
     * and the amount needed to complete the order.  However, if limit is
     * postive, no more than 'limit' units of any resource will be given.
     * @param supply the source of commodities.
     * @param limit if greater than zero, do not give more than this amount of
     * any resource.
     * @return the total number of units of resources given.
     */
    public synchronized int give(int[] supply, int limit) {
        int result = 0;
        for (int i = 0; i < request.length; i++) {
            int amt = request[i] - alloc[i];
            if (amt > supply[i]) {
                amt = supply[i];
            }
            if (limit > 0 && amt > limit) {
                amt = limit;
            }
            alloc[i] += amt;
            supply[i] -= amt;
            result += amt;
        }
        return result;
    } // grant(int[],int)

    /** Check whether this order is satified.
     * @return true if the amount allocated matches the amount requested, for
     * all resources.
     */
    public synchronized boolean satisfied() {
        for (int i = 0; i < request.length; i++) {
            if (alloc[i] != request[i]) {
                return false;
            }
        }
        return true;
    } // satisfied()

    /** Return the total amount requested.
     * @return the sum of requests for all resources.
     */
    public synchronized int size() {
        int sum = 0;
        for (int i = 0; i < request.length; i++) {
            sum += request[i];
        }
        return sum;
    } // size()

    /** Return the remaining amount requested.
     * @return the sum of request - alloc for all resources.
     */
    public synchronized int remaining() {
        int sum = 0;
        for (int i = 0; i < request.length; i++) {
            sum += request[i] - alloc[i];
        }
        return sum;
    } // remaining()

    /** Signal completion.
     * The customer who placed this order is allowed to return.
     */
    public synchronized void complete() {
        result = true;
        done = true;
        notify();
    } // complete()

    /** Reject this Order.  All resources previously granted to this Order
     * are returned to dest, and the order is completed with a failure
     * indication.
     * @param dest the place to put the revoked resources.
     */
    public synchronized void cancel(int[] dest) {
        for (int i = 0; i < request.length; i++) {
            dest[i] += alloc[i];
            alloc[i] = 0;
        }
        result = false;
        done = true;
        notify();
    } // revoke(int[])

    /** Await fulfillment of this order.
     * @return false if the order was rejected because of shutdown.
     */
    public synchronized boolean await() {
        while (!done) {
            try {
                wait();
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }
        return result;
    } // await()

    /** Turn an Order into a printable string.
     * Syntax is <code>seq:cust[alloc1/request1,...,allocn/requestn]</code>
     * where <code>seq</code> is the sequence number of this Order, 
     * <code>cust</code> is the id of the requesting Customer,
     * <code>alloci</code> is the amount of the ith commodity already allocated
     * to this order, and <code>requesti</code> is the initial request amount
     * of commodity i.
     */
    public String toString() {
        StringBuffer sb = new StringBuffer(seq + ":" + id);
        for (int i = 0; i < request.length; i++) {
            sb.append(i == 0 ? '[' : ',')
                .append(alloc[i])
                .append('/')
                .append(request[i]);
        }
        sb.append(']');
        return sb.toString();
    } // toString()

} // Order
