/* $Id: Shell.java,v 1.7 1999/11/11 13:10:00 solomon Rel solomon $ */

import java.util.*;

/** A simple command-line shell for the MiniKernel.
 *
 * <p>
 * This program displays a prompt and waits for the user to enter commands.
 * A command line consists of one or more commands, separated by ampersands
 * (&amp;).
 * A command is the name of a Java class that implements
 * <samp>public static void main(String args[])</samp>, followed by
 * zero or more arguments.
 * All the commands on the line are run in parallel.
 * The shell waits for them to finish before issuing another prompt.
 * <p>
 * The Shell terminates if it sees end-of-file (Control-D if input is coming
 * from the keyboard) or the <samp>exit</samp> command.
 * <p>
 * @see Kernel
 */
public class Shell {
    /** The main program.
     * @param args command-line arguments (ignored).
     */
    public static void main(String args[]) {
        StringBuffer sb = new StringBuffer();
        for (;;) {
            Library.output("Shell> ");
            int rc = Library.input(sb);

            if (rc == Kernel.ERROR_END_OF_FILE) {
                return;
            }
            if (rc < 0) {
                Library.output("Fatal error trying to read from console\n");
                System.exit(1);
            }
            String s = sb.toString();

            if (s.equals("exit")) {
                return;
            }

            // Split into commands separated by &
            StringTokenizer st = new StringTokenizer(s,"&");
            int commandCount = st.countTokens();
            if (commandCount == 0) {
                continue;
            }
            int[] pids = new int[commandCount];
            int nextPid = 0;

            while (st.hasMoreTokens()) {
                String command = st.nextToken();

                // Split each command by spaces
                StringTokenizer cst = new StringTokenizer(command);
                if (!cst.hasMoreTokens()) {
                    // empty command
                    pids[nextPid++] = -1;
                    continue;
                }

                String program = cst.nextToken();
                String[] progArgs = new String[cst.countTokens()];
                for (int i = 0; cst.hasMoreTokens(); ) {
                    progArgs[i++] = cst.nextToken();
                }

                int pid = Library.exec(program,progArgs);
                if (pid < 0) {
                    Library.output("Shell: Error executing " + command
                        + ":  " + Library.errorMessage[-pid] + "\n");
                }
                pids[nextPid++] = pid;
            }

            // Wait for all the processes to complete
            for (int i = 0; i < commandCount; i++) {
                if (pids[i] < 0) {
                    continue;
                }
                if (Library.join(pids[i]) != 0) {
                    Library.output("Error waiting for process " + pids[i]
                        + " to complete\n");
                }
            }
        } // for(;;)
    } // main
} // class Shell
