import java.util.*;

public class CachedScheduler extends Scheduler {

    // Disk Cache
    protected CacheBlock[] cache;
    protected int ctail;
    protected Vector freeBlockWaitingList;

    // Disable default constructor
    protected CachedScheduler() {
	super();
    }

    public CachedScheduler(int cacheSize, Disk disk) { //initalize the CachedScheduler
	// initialize the Scheduler
	super(cacheSize, disk);
	// initialize the cache
	cache = new CacheBlock[cacheSize];
	for (int i=0; i < cacheSize; i++) {
	    byte[] cacheBuffer=new byte[disk.BLOCK_SIZE];
	    cache[i] = new CacheBlock(cacheBuffer);
	}
        //setup the FIFO queue pointers
	ctail = 0;
	// setup the free block waiting list
	freeBlockWaitingList = new Vector();
    }
    
    public void requestRead(UserRequest userReq) { //note it is not sychronized
	CacheBlock freeBlock;
	for (int i=0; i < cache.length ; i++) {
	    if (cache[i].blockNumber == userReq.blockNumber) {
		freeBlock = doBlockInCache(cache[i]);
		System.out.println("CS: block "+userReq.blockNumber+" is already in cache");
		System.out.println("           with status="+cache[i].status);
		//finish copy the cache to user
		System.arraycopy(freeBlock.buffer, 0, userReq.buffer, 0, freeBlock.buffer.length);
		// wake the user Request up
		freeBlock.signalCompletion(); //tell other process waiting on the cache block
		userReq.signalCompletion();   //tell user the request is finish
		return;
	    }
	}
	
	// the doFindCleanBlock will find a clean block in cache for replacement
	// moreover, it will set the "new" blockNumber and status (according to the userRequest
	// of the clean block b4 return
	
	freeBlock = doFindCleanBlock(userReq);
	// from the above a clean block is get
	// request a read from disk (the instructions have to be atomic), thus inside 
	// a synchronized method
	if (freeBlock.blockNumber != userReq.blockNumber) {
	    waitForReadDisk(freeBlock, userReq);
	}
	// need block in freeBlock now
	
	//finish copy the cache to user
	System.arraycopy(freeBlock.buffer, 0, userReq.buffer, 0, freeBlock.buffer.length);
	// wake the user Request up
	freeBlock.signalCompletion(); //tell other process waiting on the cache block
	userReq.signalCompletion();   //tell user the request is finish
	return;
    }

    public void requestWrite(UserRequest userReq) {
	CacheBlock freeBlock;
	for (int i=0; i < cache.length ; i++) {
	    if (cache[i].blockNumber == userReq.blockNumber) {
		freeBlock = doBlockInCache(cache[i]);
		System.out.println("CS: block "+userReq.blockNumber+" is already in cache");
		System.out.println("           with status="+cache[i].status);
		//finish copy the user data to cache
		atomicWrite(userReq, freeBlock);
		// wake the user Request up
		freeBlock.signalCompletion(); //tell other process waiting on the cache block
		userReq.signalCompletion();   //tell user the request is finish
		return;
	    }
	}
	
	// the doFindCleanBlock will find a clean block in cache for replacement
	// moreover, it will set the "new" blockNumber and status (according to the userRequest
	// of the clean block b4 return
	
	freeBlock = doFindCleanBlock(userReq);
	// from the above a clean block is get
	// need block in freeBlock now
	
	//finish copy the user data to cache
	atomicWrite(userReq, freeBlock);
	// wake the user Request up
	freeBlock.signalCompletion(); //tell other process waiting on the cache block
	userReq.signalCompletion();   //tell user the request is finish
	return;
    }
    
    public synchronized void atomicWrite(UserRequest uReq, CacheBlock cBlock){
	System.arraycopy(uReq.buffer, 0, cBlock.buffer, 0, cBlock.buffer.length);
	if (cBlock.status == CacheBlock.UNUSED) {
	    cBlock.status = CacheBlock.CLEAN;
	} else {
	    cBlock.status = CacheBlock.DIRTY;
	}
    }

    public synchronized void waitForReadDisk(CacheBlock cBlock, UserRequest uReq){
	// might not be need depends on how doFindCleanBlock impletement
	//cBlock.status = CacheBlock.READING;
	//cBlock.blockNumber = uReq.blockNumber;
	//---------------------------------
	Request req = super.requestRead(cBlock.blockNumber, cBlock.buffer);
	req.waitForCompletion(); //wait for the disk request complete
	//copy the data to cacheblock
	System.arraycopy(req.buffer, 0, cBlock.buffer, 0, cBlock.buffer.length);
	cBlock.status = CacheBlock.CLEAN;
	return;
    }

    public synchronized CacheBlock doBlockInCache(CacheBlock cBlock) {
	if (cBlock.status == CacheBlock.READING || cBlock.status == CacheBlock.WRITING) {
	    cBlock.waitCount++;
	    System.out.println("CachedSchedular : DBIC busy block " + cBlock.blockNumber+" in Cache. Status = " + cBlock.status ); 
	    cBlock.waitForCompletion();
	    System.out.println("CachedSchedular : DBIC block " + cBlock.blockNumber+" in Cache. Status = " + cBlock.status + " . Wait Completed ");
	    cBlock.waitCount--;
	} else {
	    System.out.println("CachedSchedular : DBIC ready block " + cBlock.blockNumber+" in Cache. Status =" + cBlock.status ); 
	    // the cacheblock is either clean or dirty which is valid for read, so do nothing
	}
	return cBlock;
    }

    public synchronized CacheBlock doFindCleanBlock(UserRequest uReq) {
	System.out.println("CachedScheduler: finding free block in cache");
	for (;;) {
	    int startPos=ctail;
	    int seekPos=ctail;
	    for (;;) {
		//case 1 block is clean or unused
		if (cache[seekPos].status == CacheBlock.CLEAN ||
		    cache[seekPos].status == CacheBlock.UNUSED) {
		    cache[seekPos].blockNumber = uReq.blockNumber;
		    if (uReq.status == UserRequest.READ) {
			cache[seekPos].status = CacheBlock.READING;
		    }
		    if (uReq.status == UserRequest.WRITE) {
			
			/*
			  for (int i=0; i<cache.length; i++) {
			  if (cache[i].blockNumber == uReq.blockNumber) {
			  CacheBlock existBlock = doBlockInCache(cache[i]);	    
			  return existBlock;
			  } else {
			  
			  }
			  }
			*/
	    // this "writing flag" is not a write from buffer to disk
	    // but a write from user to buffer. (1) no disk op
	    // (2) will be set to its correct value "dirty" by caller
			cache[seekPos].status = CacheBlock.WRITING;
		    }
		    System.out.println("CachedScheduler: found block #"+
				       seekPos+" for request "+
				       uReq.toString());
		    ctail++;
		    if (ctail == cache.length ) ctail = 0;
		    return cache[seekPos];
		}
		
		//case 2 block is dirty (need cleaning, enqueue write to disk)
		if (cache[seekPos].status == CacheBlock.DIRTY) {
		    cache[seekPos].status = CacheBlock.WRITING;
		    Request evictReq = super.requestWrite(cache[seekPos].blockNumber,
							  cache[seekPos].buffer);
		    // the request handle will be discarded &
		    // CachedScheduler will recapture this event when its 
		    // completion causes an interrupt.	    
		}

		//case 3 block is reading or writing, no need to check as implicitly implemented

		// If case is (Dirty=>writing, reading, writing) => check next.
		seekPos++;
		if (seekPos == cache.length) seekPos=0;
		if (seekPos == startPos) break;
	    }

	    System.out.println("CachedScheduler: cannot find any free block:"+
			       " sleeping (wait on An avaliable clean block)");
		// wait for at least one block is clean
		// need sonething here ******
	    freeBlockWaitingList.add(uReq);
	    try {
		for (;;) {
		    wait();
		    if (uReq.cacheAllocated != -1) break;
		}
	    } catch (InterruptedException ex) {
		ex.printStackTrace();
	    }
	    //    uReq.waitForCacheFree();
	    
	    // There are two cases: 
	    // (1) there is a clean cache block
	    // (2) someone brought in the physical block
	    // so, determine which one.
	   
	    // After sleeping, check if the request is already complete
	    // by someone else (need to check cache)
	    for (int i=0; i<cache.length; i++) {
		if (cache[i].blockNumber == uReq.blockNumber) {
		    CacheBlock existBlock = doBlockInCache(cache[i]);	    
		    return existBlock;
		}
	    }
	    // if the block is not already in the cache (which should be the normal case)
	    if (cache[uReq.cacheAllocated].status == CacheBlock.CLEAN) {
		// return the clean available block
		return cache[uReq.cacheAllocated];
	    } else {
		// we have a problem; we don't know why we're awaken.
		System.out.println("CachedScueduler: Immature leave from "+
				   "doFindCleanBlock()");
	    }
	}
    }

    public synchronized void interrupt() {
	System.out.println("Disk interrupt entered CS.");
	notifyAll();
	if (currentRequest.status == Request.READ) {
	    // Make the corresponding block clean
	    // since we haven't keep track of cache number, 
	    // we'll do a search
	    int blockNumber = currentRequest.blockNumber;
	    int cacheNumber = -1;
	    for (int i=0; i<cache.length; i++) {
		if (cache[i].blockNumber == blockNumber) {
		    cache[i].status = CacheBlock.CLEAN;
		    cacheNumber = i;
		    System.out.println("CS: cache #"+cacheNumber+
				       " with block "+blockNumber+
				       " is made clean.");
		}
	    }
	    if (cacheNumber == -1) {
		// we have a problem here, someone accessed disk
		// with a buffer outside the cache
		System.out.println("CacheScheduler: Error in interrupt");
	    }
	    super.interrupt();
	    return;
	} else if (currentRequest.status == Request.WRITE) {
	    // Make the corresponding block clean
	    // since we haven't keep track of cache number, 
	    // we'll do a search
	    int blockNumber = currentRequest.blockNumber;
	    int cacheNumber = -1;
	    for (int i=0; i<cache.length; i++) {
		if (cache[i].blockNumber == blockNumber) {
		    cache[i].status = CacheBlock.CLEAN;
		    cacheNumber = i;
		}
	    }
	    if (cacheNumber == -1) {
		// we have a problem here, someone accessed disk
		// with a buffer outside the cache
		System.out.println("CacheScheduler: Error in interrupt");
	    }
	    // That cache block cleaned, see if anyone waiting
	    // for that cache block's data
	    CacheBlock cBlock = cache[cacheNumber];
	    if (cache[cacheNumber].waitCount > 0) {
		cache[cacheNumber].signalCompletion();
		System.out.println("CachedScheduler: "+
				   "write "+cBlock.blockNumber+
				   " completed. waitCount="+cBlock.waitCount);
	    } else {
		System.out.println("CachedScheduler: "+
				   "evict phys block "+cBlock.blockNumber+
				   " completed.");
	        // no one waiting..   check freeBlockWaitingList
		if (freeBlockWaitingList.size()>0) {
		    UserRequest uReq = (UserRequest) freeBlockWaitingList.remove(0);
		    uReq.signalCacheFree(cacheNumber);
		}
	    }
	    super.interrupt();
	}
    }

    public void shutdown() {
    }
    

}




