import java.util.*;

public class FileSystem {

    // Values for diskStatus
    public static final int FORMATTED = 0;
    public static final int UNFORMATTED = -1;
    public static final int UNKNOWN_MEDIA = -2;

    // Values for "requesting special treatment" for private operations
    // 
    public static final int READ = 20;
    public static final int WRITE = 21;

    // Values for "dereference" for handling missing indirect blocks.
    public static final int READ_INODE = 30;
    public static final int READ_DATA = 31;
    public static final int WRITE_INODE = 40;
    public static final int WRITE_DATA = 41;

    protected FastDisk disk;
    protected int diskStatus;

    protected int numBlocks;
    protected int blockSize;
    protected int maxFiles;

    protected int freeMapStart;
    protected int freeMapSize;
    protected int freeMapStop;  // the last block (inclusive) in free Map.
    protected int rootBlockStart;


    protected int inodePerBlock;
    protected int ptrPerBlock;
    protected int bitPerBlock;

    // Number of D-type blocks in the root -> tree -> inode struture. 
    protected int treeDepth;

    // Number of file data blocks supported by each level of indirection
    // (these values are not accumulative)
    // [0] = 11 (stored within inode, offset 4)
    // [1] = 1 (in inode, offset 26) * ptrPerBlock 
    // [2] = 1 (in inode, offset 28) * (ptrPerBlock ** 2)
    // [3] = 1 (in inode, offset 30) * (ptrPerBlock ** 3)
    protected int[] fileBlocksPerLevel = new int[4];
    
    public FileSystem() {
	System.out.println("Error: (FileSystem) Unsupported constructor.");
    }
	
    public FileSystem(FastDisk disk) {
	this.disk = disk;
	// Attempt to read superblock.
	byte[] buffer = new byte[disk.BLOCK_SIZE];
	disk.read(0, buffer);
	int storedDiskSize = unpackShort(buffer, 0);
	int storedMaxFiles = unpackShort(buffer, 2);
	if (storedDiskSize == 0) {
	    diskStatus = UNFORMATTED;
	    System.out.println("Disk registered as unformatted."+
			       " device-reported byte size "+
			       disk.DISK_SIZE * disk.BLOCK_SIZE);
	} else if (storedDiskSize == disk.DISK_SIZE) {
	    diskStatus = FORMATTED;
	    System.out.println("Disk mounted, byte size "+
			       storedDiskSize * disk.BLOCK_SIZE);
	} else {
	    diskStatus = UNKNOWN_MEDIA;
	    System.out.println("Disk is invalid; not mounted."+
			       " device-reported byte size "+
			       disk.DISK_SIZE * disk.BLOCK_SIZE);
	}
	if (diskStatus == FORMATTED) {
	    // Physical parameters
	    this.numBlocks = storedDiskSize;
	    this.blockSize = disk.BLOCK_SIZE;
	    this.maxFiles = storedMaxFiles;
	    // Compute rest of file system parameters.
	    // (const values, free map start & stop, root block number)
	    computeDiskParams(numBlocks, blockSize, maxFiles);
	    System.out.println(printFreeMap());
	}
    }

    protected void computeDiskParams(int numBlocks,
				   int blockSize,
				   int maxFiles) 
    {
	// Constants used in calculations
	this.inodePerBlock = blockSize >> 5; // div 32
	this.ptrPerBlock = blockSize >> 1;   // div 2
	this.bitPerBlock = blockSize << 3;  // mul 8
	// Size and offset of Free Map, root block
	this.freeMapStart = 1;
	this.freeMapSize = (numBlocks / bitPerBlock) +
	    ( ((numBlocks % bitPerBlock) == 0) ? (0) : (1) );
	this.freeMapStop = freeMapSize;
	this.rootBlockStart = freeMapStop + 1;
	
    findTreeDepth:
	{ // find depth of root directory 
	    int structLimit=inodePerBlock;
	    for (int depth=0; depth<numBlocks; depth++) {
		if (maxFiles <= structLimit) {
		    this.treeDepth = depth;
		    break findTreeDepth;
		} else {
		    structLimit *= ptrPerBlock;
		}		
	    }
	}
	// Since according to specification, root block cannot 
	// be a inode-block, so it has to be an indirect block
	// (so minimum level = 1, for 0 <= maxFiles <= D*I)
	if (treeDepth == 0) 
	    treeDepth = 1;
	
    findFileBlocksPerLevel:
	{   // find out how many file blocks can fit in level (0,1,2)
	    // which is used to determine which level 
	    // a file offset belongs to
	    fileBlocksPerLevel[0] = 11;
	    fileBlocksPerLevel[1] = fileBlocksPerLevel[0] + ptrPerBlock;
	    fileBlocksPerLevel[2] = fileBlocksPerLevel[1] +
		ptrPerBlock * ptrPerBlock;
	    fileBlocksPerLevel[3] = fileBlocksPerLevel[2] +
		ptrPerBlock * ptrPerBlock * ptrPerBlock;    
	}
    }


      /** Performs indirect block dereferencing (given a starting point and 
       *  an offset, reads the pointer stored at that offset, and allocate
       *  a new block if necessary.
       *
       *  @param operation READ or WRITE.
       *  @param parent    the block number of a Indirect Block.
       *  @param ptrIndex  The index of that pointer in the Indirect Block
       *  @return The value of the pointer stored there.
       */
    public int dereference(int operation, int parent, int ptrIndex) {
	if (operation != READ && operation != WRITE) {
	    System.out.println("Error dereference(*): invalid operation"+
			       " code ("+operation+")"); 
	    return -1;
	}
	if (parent < rootBlockStart || parent >= numBlocks) {
	    System.out.println("Error dereference(*): invalid starting"+
			       " block ("+parent+")"); 
	    return -1;
	}
	if (ptrIndex < 0 || ptrIndex >= ptrPerBlock) {
	    System.out.println("Error dereference(*): invalid ptrIndex ("+
			       ptrIndex+")");
	    return -1;
	}
	byte[] buffer = new byte[blockSize];
	// Read in the indirect block.
	disk.read(parent, buffer);
	// access the i-th pointer.
	int result = unpackShort(buffer, ptrIndex*2);
	// If the pointer is null, and the caller wants to allocate it
	if (result == 0 && operation == WRITE) {
	    // Allocate a block and set pointer to point to it
	    result = getNewBlock();
	    if (result <= 0) {
		// Cannot get a new block: return error.
		return -1;
	    } 
	    packShort(result, buffer, ptrIndex*2);
	    // The indirect block is updated to the disk.
	    disk.write(parent, buffer);
	}
	System.out.println("block["+parent+"].ptr["+ptrIndex+"]="+result);
	return result;
    }


    /** 
     *  Retrieve an inode record (32-byte, representing a file) from
     *  the file system, given the file number. The "operation"
     *  specifies what to do when the inode block doesn't exist:
     *  when it's a read, the file tree will be left as is; 
     *  when it's a write, the file will be created and any necessary
     *  inode blocks (etc) will be allocated.
     *
     *  @param fileNumber  the "numerical name" of the file
     *  @param operation   specifies what to do when the file doesn't exist
     *  @param ifield      Stores the return value (the 32-byte inode record) 
     *
     *  @return  an Inode object if the file exists or is just created
     *           returns null otherwise.
     */
    public Inode namei(int fileNumber, int operation) {
	byte[] buffer = new byte[blockSize];
	byte[] inodeBuf = new byte[32];   // temporary buffer for search
	int[] index = new int[4];

	if (fileNumber < 0 || fileNumber >= maxFiles) {
	    return null;
	}

	if (operation != READ && operation != WRITE) {
	    return null;
	}

	// Find out the "record number", or the "offset", at each level
	// in the file tree.
    computeIndexForEachLevel:      
	{
	    int tmp = fileNumber;
	    index[0] = tmp % inodePerBlock;
	    tmp /= inodePerBlock;
	    index[1] = tmp % ptrPerBlock;
	    tmp /= ptrPerBlock;
	    index[2] = tmp % ptrPerBlock;
	    tmp /= ptrPerBlock;
	    index[3] = tmp; /* % ptrPerBlock */ ;
	}

	int currentBlock = rootBlockStart;

	// recurse into the i-block that holds the inode data.
	// (do not look into the individual fields in i-block)
	for (int i = treeDepth; i > 0; i--) {
	    int p = dereference(READ, currentBlock, index[i]);
	    if (p > 0) {
		// Next level is allocated
		currentBlock = p;
	    } else {
		// Next level is not allocated
		if (operation == READ) {
		    // If read, return to caller as "node not found".
		    return null;
		} 
		else if (operation == WRITE) {
		    // If write, allocate this block.
		    p = dereference(WRITE, currentBlock, index[i]);
		    // System.out.println("namei allocated: "+p);

		    if (p <= 0) { // cannot allocate new block
			return null;
		    }
		    // if next level is another indirect block, fill zero.
		    // if next level is inode block, fill sizes with -1.
		    if (i > 1) {
			// block is already zero-filled by getNewBlock;
			// called by dereference(,,).
		    } else /* (i == 1) */ {
			initializeInodeBlock(p);
		    }
		    currentBlock = p;
		}
	    }
	    // Finish descending into next level.
	}
	// Now currentBlock points to an Inode block.
	disk.read(currentBlock, buffer);
	System.arraycopy(buffer, index[0]*32, inodeBuf, 0, 32);

	// If the Inode is unused, and a WRITE is requested,
	// the inode will be initialized. (the file is created.)
	// (size = 0, pointers = 0)
	if (operation == WRITE &&
	    unpackInt(inodeBuf, 0) == -1) {
	    fillByte(inodeBuf, 0);
	    System.arraycopy(inodeBuf, 0, buffer, index[0]*32, 32);
	    disk.write(currentBlock, buffer);
	}
	System.out.println("block["+currentBlock+"].inode["+index[0]+
			   "]= (inumber="+fileNumber+"; size="+
			   unpackInt(inodeBuf, 0)+")");

	// When finished, return 0 if file exists or is created.
	if (unpackInt(inodeBuf, 0) >= 0) {
	    Inode inode = new Inode();
	    inode.unpack(inodeBuf, 0); // inodeBuf
	    inode.inumber = fileNumber;
	    inode.inodeBlock = currentBlock;
	    inode.inodeBlockOffset = index[0];
	    return inode;
	}
	// Otherwise, return null to indicate that file does not exist.
	return null;
    }

    public int storeInode(Inode inode) {
	byte[] buffer = new byte[blockSize];
	byte[] inodeBuf = new byte[32];   // temporary buffer for search
	int[] index = new int[3];
	if (inode == null) {
	    System.out.println("Error (int) storeInode(Inode) : "+
			       "null pointer in parameter");
	    return -1;
	}
	
	if (inode.inumber < 0 || inode.inumber >= maxFiles) {
	    System.out.println("Error (int) storeInode(Inode) : "+
			       " invalid (int) Inode.inumber");
	    return -1;
	}
	
	// Find out the "record number", or the "offset", at each level
	// in the file tree.
    computeIndexForEachLevel:      
	{
	    int tmp = inode.inumber;
	    index[0] = tmp % inodePerBlock;
	    tmp /= inodePerBlock;
	    index[1] = tmp % ptrPerBlock;
	    tmp /= ptrPerBlock;
	    index[2] = tmp /* % ptrPerBlock */ ;
	}

	int currentBlock = rootBlockStart;

	// recurse into the i-block that holds the inode data.
	// (do not look into the individual fields in i-block)
	for (int i = treeDepth; i > 0; i--) {
	    int p = dereference(READ, currentBlock, index[i]);
	    if (p > 0) {
		// Next level is allocated
		currentBlock = p;
	    } else {
		// allocate blocks for any missing ones.
		p = dereference(WRITE, currentBlock, index[i]);
		// System.out.println("namei allocated: "+p);
		
		if (p <= 0) { // cannot allocate new block
		    return -1;
		}
		// if next level is another indirect block, fill zero.
		// if next level is inode block, fill sizes with -1.
		if (i > 1) {
		    // block is already zero-filled by getNewBlock;
		    // called by dereference(,,).
		} else /* (i == 1) */ {
		    initializeInodeBlock(p);
		}
		currentBlock = p;
	    }
	    // Finish descending into next level.
	}
	// Now currentBlock points to an Inode block.
	disk.read(currentBlock, buffer);
	
	// Save this inode record into the Inode-block.
	inode.pack(buffer, index[0]);
	disk.write(currentBlock, buffer);
	System.out.println("block["+currentBlock+"].inode["+index[0]+
			   "]= (inumber="+inode.inumber+"; size="+
			   inode.length+")");	
	return 0;	
    }

    /****************** Supported disk actions ********************/

    public int format(int maxFiles) {
	// if maxFiles too small, exit with error
	if (maxFiles <= 0) {
	    return -1;
	}

	// Block buffer.
	byte[] buffer = new byte[disk.BLOCK_SIZE];
	
	// Compute major disk / file system parameters
	this.numBlocks = disk.DISK_SIZE;
	this.blockSize = disk.BLOCK_SIZE;
	this.maxFiles = maxFiles;

	computeDiskParams(this.numBlocks, this.blockSize, this.maxFiles);

	System.out.println("Formatting disk:\n"+
			   "numBlocks="+numBlocks+"\n"+
			   "blockSize="+blockSize+"\n"+
			   "maxFiles="+maxFiles+"\n"+
			   "freeMapStart="+freeMapStart+"\n"+
			   "freeMapStop="+freeMapStop+"\n"+
			   "freeMapSize="+freeMapSize+"\n"+
			   "rootBlockStart="+rootBlockStart+"\n"+
			   "inodePerBlock(I)="+inodePerBlock+"\n"+
			   "ptrPerBlock(D)="+ptrPerBlock+"\n"+
			   "bitPerBlock="+bitPerBlock+"\n"+
			   "treeDepth="+treeDepth+"\n");

	// Initialize superblock.
	fillByte(buffer, 0); // optional
	packShort(numBlocks,buffer, 0);
	packShort(maxFiles, buffer, 2);
	disk.write(0, buffer);

	// Initialize free map.
    initializeFreeMap:
	{
	    for (int bn = freeMapStart; bn <= freeMapStop; bn++) {
		fillByte(buffer, 0);   // all 0's, "unused"
		for (int index = 0; index < bitPerBlock; index++) {
		    int blockNumber = (bn-freeMapStart) * blockSize + index;
		    // for system area and unmapped area, set as "used".
		    if (blockNumber <= rootBlockStart ||
			blockNumber >= numBlocks) {
			// print diagnostic message
			if (blockNumber <= rootBlockStart) {
			    System.out.println("block "+blockNumber+
					       " marked as used.");
			} else if (blockNumber == numBlocks) {
			    System.out.println("blocks "+blockNumber+
					       " and above"+
					       " marked as unavailable.");
			}
			// set as "used"
			buffer[index>>3] |= (1 << (index & 7));
			
		    }
		}
		disk.write(bn, buffer);
	    }
	}
    
	// Initialize root block 
	// if depth=0, fill inode with -1.
	// if depth>0, fill with zero (i.e. indirect blocks not allocated)
	fillByte(buffer, 0);
	if (treeDepth==0) {
	    for (int index = 0; index < inodePerBlock; index++) {
		packInt(-1, buffer, index*32);
	    }
	} else /* (treeDepth > 0) */ {
	    /* fill with zero - already done*/
	}
	disk.write(rootBlockStart, buffer);

	// return to caller
	this.diskStatus = FORMATTED;


	// Report disk usage.
	System.out.println(printFreeMap());
	return 0;
    }
    

    // Returns with an available new block, and marks the block
    // as "used" in the free map. The block is zero-filled.
    //
    //
    //
    public int getNewBlock() {
	// Check if disk is formatted.
	if (diskStatus != FORMATTED) 
	    return -1;

	byte[] buffer = new byte[blockSize];
	// search (from first to last)
	// Reserved areas and invalid areas are marked as "used (=1)"
	// All bits in the freeMap (even those corresponding to blocks
	// beyond capacity) may be searched.
	//
	int blockNumber = -1;
    searchFreeMap:
	{
	    for (int bn = freeMapStart; bn <= freeMapStop; bn++) {
		disk.read(bn, buffer);
		for (int index = 0; index < bitPerBlock; index++) {
		    if ( ((buffer[index>>3] >> (index & 7)) & 1) == 0 ) {
			buffer[index>>3] |= (1 << (index & 7));
			disk.write(bn, buffer);
			blockNumber= (bn-freeMapStart) * blockSize + index;
			break searchFreeMap; 
		    }
		}    
	    }
	}
	if (blockNumber == -1) {
	    System.out.println("Out of free blocks.");
	    return -1;
	}
	if (blockNumber <= rootBlockStart) {
	    System.out.println("Error in FS.getNewBlock():"+
			       " reserved block marked as"+
			       " free.");
	    return -1;
	}
	if (blockNumber >= numBlocks) {
	    System.out.println("Error in FS.getNewBlock():"+
			       " out-of-bound block flags"+
			       " marked as free.");
	    return -1;
	}
	// Zero-fill the block
	// ### "byte[] buffer" will now be reclaimed. ###
	fillByte(buffer, 0);
	disk.write(blockNumber, buffer);

	// return block number.
	return blockNumber;
    }
    
    public int releaseBlock(int blockNumber) {
	// Check if disk is formatted.
	if (diskStatus != FORMATTED) 
	    return -1;

	// Validate block number
	if (blockNumber <= rootBlockStart) {
	    System.out.println("Error in FS.releaseBlock(int):"+
			       " attempt to release reserved block.");
	    return -1;
	}
	if (blockNumber >= numBlocks) {
	    System.out.println("Error in FS.releaseBlock(int):"+
			       " block number too large.");
	    return -1;
	}
	byte[] buffer = new byte[blockSize];
	// The block number containing that flag
	int bn = (blockNumber / bitPerBlock) + freeMapStart;
	int index = blockNumber % bitPerBlock;
	
	disk.read(bn, buffer);

	if ((buffer[index>>3] & (1 << (index & 7))) == 0) {
	    System.out.println("Error in FS.releaseBlock(int):"+
			       " attempt to release unused block.");
	    return -1;
	}

	buffer[index>>3] &= (~( 1 << (index & 7) ));

	disk.write(bn, buffer);

	return 0;
    }


    public void initializeInodeBlock(int blockNumber) {
	byte[] buffer = new byte[blockSize];
	for (int i=0; i < inodePerBlock; i++) {
	    packInt(-1, buffer, i*32);
	}
	disk.write(blockNumber, buffer);
    }

    public String printFreeMap() {
	byte[] buffer = new byte[blockSize];
	StringBuffer sbuf = new StringBuffer("Free Map: \n");
	int available = 0;
    iterateFreeMap:
	for (int bn=freeMapStart; bn <= freeMapStop; bn++) {
	    disk.read(bn, buffer);
	    for (int index=0; index < bitPerBlock; index++) {
		int blockNumber = (bn-freeMapStart)*bitPerBlock + index;
		if ((blockNumber % 64) == 0)
		    sbuf.append(blockNumber+": ");
		int flag=(buffer[index>>3] >> (index & 7)) & 1;
		sbuf.append(flag+"");
		if (flag == 0) 
		    available++;
		if (blockNumber == numBlocks-1) {
		    sbuf.append("\n");
		    break iterateFreeMap;
		}
		else if ((blockNumber % 64) == 63)
		    sbuf.append("\n");
		else if ((blockNumber % 16) == 15)
		    sbuf.append("|");
		else if ((blockNumber % 16) == 7)
		    sbuf.append(".");
	    }
	}
	sbuf.append("End of free map\n");
	sbuf.append("Number of free blocks: "+available+"\n");
	return sbuf.toString();
    }

    /** Store a 16-bit integer into a byte array.
     * @param n the integer to be stored
     * @param buf the byte array into which it should be stored
     * @param offset the index of the first byte to be modified
     */
    protected static void packShort(int n, byte[] buf, int offset) {
        buf[offset] = (byte) (n >> 8);
        buf[offset+1] = (byte) n;
    }

    /** Convert a field in a byte array to a 16-bit integer.
     * @param buf the byte array containing the data.
     * @param offset the location in the array where the data starts.
     * @return the short integer value.
     */
    protected static int unpackShort(byte[] buf, int offset) {
        return ((buf[offset] & 0xff) << 8) + (buf[offset+1] & 0xff);
    }

    
    /** Store a 32-bit integer into a byte array.
     * @param n the integer to be stored
     * @param buf the byte array into which it should be stored
     * @param offset the index of the first byte to be modified
     */
    protected static void packInt(int n, byte[] buf, int offset) {
	buf[offset] = (byte) (n >> 24);
        buf[offset+1] = (byte) (n >> 16);
        buf[offset+2] = (byte) (n >> 8);
        buf[offset+3] = (byte) n;
    }

    /** Convert a field in a byte array to a 32-bit integer.
     * @param buf the byte array containing the data.
     * @param offset the location in the array where the data starts.
     * @return the integer value.
     */
    protected static int unpackInt(byte[] buf, int offset) {
        return 
	    ((buf[offset] & 0xff) << 24) + 
	    ((buf[offset+1] & 0xff) << 16) +
	    ((buf[offset+2] & 0xff) << 8) +
	    ((buf[offset+3] & 0xff));
    }

    protected static void fillByte(byte[] buf, int value) {
	for (int i=0; i<buf.length; i++) {
	    buf[i] = (byte) value;
	}
    }

    protected static String printIntArray(int[] array) {
	StringBuffer buf=new StringBuffer();
	for (int i=0; i<array.length; i++) {
	    buf.append(array[i]+"");
	    if (i != (array.length-1)) {
		buf.append(",");
	    }
	    if ((i % 8) != 7) {
		buf.append(" ");
	    } else {
	        buf.append("\n");
	    }
	}
	return buf.toString();
    }

    /*************************************************************************
     * Block access within a file
     * takes the operation going to perform with the request block, 
     * an Inode and a block-offset of a block within the file 
     * (lets say block A)
     *  return a pointer (short) to block A 
     * (i.e. return the block number of block A in disk)
     * @param operation determine what to do if the require block do no exist
     * @param inode the inode reference to the request block
     * @param offset the block offset with in the file of the request block
     * 
     * @return 0 if operation = READ and block not exist
     *                      (caller will handle to output a zeroed buffer)
     * @return -1 if there is an error occur
     *
     * assume offset within the possible reference range (handle by caller)
     *************************************************************/
    
    public int getBlockNum(int operation, Inode inode, int offset){
	if (operation != READ && operation != WRITE) {
	    System.out.println("getBlockNum Error: Invalid Operation code (" +
			       operation +")");   
	    return -1;
	}
	
	if (offset < fileBlocksPerLevel[0]) { // lvl 0
	    
	    // check if the block exists
	    if (inode.direct[offset] == 0 && operation == WRITE){ 
		// block does not exists, allocate it in the disk
		// not return error untill cannot allocate new block 
		// (i.e. disk full)
		
		// allocate the new block
		int newBlkNum = getNewBlock();
		if (newBlkNum != -1){
		    inode.direct[offset] = (short) newBlkNum;
		    storeInode(inode);
		    return newBlkNum;
		}
		else{ // cannot allocate new free block (disk full), 
		    // propagate the error
		    return -1;
		}
	    }
	    else {
		// operation == READ or block exist
		return inode.direct[offset];
	    }
	}
	else /*lvl 1*/ if ( (offset >= fileBlocksPerLevel[0]) &&
			    (offset < fileBlocksPerLevel[1]) ) { 
	    
	    /* 3 case here (1) indirect1 = 0 && WRITE
	                   (2) indirect1 != 0
	                   (3) indirect1 = 0 && READ
	    */
	    
	    int adjOffset = offset - fileBlocksPerLevel[0];
	    
	    // case 1
	    if (inode.indirect1 == 0 && operation == WRITE) {
		
		// allocate new block for indirect1
		int newBlkNum  = getNewBlock(); 
		if (newBlkNum < 0 ){ // when no free block avaliable error !
		    System.out.println("getBlockNum Error: Disk is Full, no new free block. Operation: "+ operation);
		    return -1;
		} else {
		    inode.indirect1 = (short) newBlkNum;
		    storeInode(inode);
		}
	    }
	    
	    // case 3
	    if (inode.indirect1 == 0 && operation == READ)
		// tell the caller it is case 3 by return zero (the pointer of indirect1)
		// the caller will handle the case (3)
		return inode.indirect1;
	    else { // case 1 and 2
		
		// find data block (or get a free data block and reference to it in case 1)
		return dereference(operation, inode.indirect1, adjOffset);
	    }
	}
	else /* lvl 2*/ if ( (offset >= fileBlocksPerLevel[1]) &&
			     (offset < fileBlocksPerLevel[2]) ) {
	    
	    /* 3 case here (1) indirect2 = 0 && WRITE
	                   (2) indirect2 != 0
	                   (3) indirect2 = 0 && READ
	    */
	    
	    int adjOffset = offset - fileBlocksPerLevel[1];
	    
	    // lvl1 Offset
	    int lvl1Offset = adjOffset % ptrPerBlock;
	    
	    adjOffset /= ptrPerBlock;
	    
	    // top lvl (lvl0) Offset
	    int toplvlOffset = adjOffset;
	    
	    
	    // case 1
	    if (inode.indirect2 == 0 && operation == WRITE) {
		
		// allocate new block for indirect2
		int newBlkNum  = getNewBlock(); 
		if (newBlkNum < 0 ){ // when no free block avaliable error !
		    System.out.println("getBlockNum Error: Disk is Full, no new free block. Operation: "+ operation);
		    return -1;
		}
		else { 
		    inode.indirect2 = (short) newBlkNum; 
		    storeInode(inode);
		}
	    }
	    
	    // case 3
	    if (inode.indirect2 == 0 && operation == READ)
		// tell the caller it is case 3 by return zero (the pointer of indirect2)
		// the caller will handle the case (3)
		return inode.indirect2;
	    else { // case 1 and 2
		
		// get the lvl1 reference indirect block 
		// note: indirect2 is the top lvl (lvl0)
		int lvl1ref =  dereference(operation, inode.indirect2, toplvlOffset);
		
		// If this block is unallocated, and operation == WRITE,
                // it will be allocated.
                if (lvl1ref == 0 && operation == READ) {
                    return 0;
                }

		// find or add data block in appropriate position
		return dereference(operation, lvl1ref, lvl1Offset);
	    }
	}
	else /* lvl 3*/ if ( (offset >= fileBlocksPerLevel[2]) &&
			     (offset < fileBlocksPerLevel[3]) ) {
	    
	    /* 3 case here (1) indirect2 = 0 && WRITE
	                   (2) indirect2 != 0
	                   (3) indirect2 = 0 && READ
	    */
	    
	    int adjOffset = offset - fileBlocksPerLevel[1];
	    
	    // lvl2 Offset
	    int lvl2Offset = adjOffset % ptrPerBlock;
	    
	    adjOffset /= ptrPerBlock;
	    
	    // lvl1 Offset
	    int lvl1Offset = adjOffset % ptrPerBlock;
	    
	    adjOffset /= ptrPerBlock;
	    
	    // top lvl (lvl0) Offset
	    int toplvlOffset = adjOffset;
    
	    // case 1
	    if (inode.indirect3 == 0 && operation == WRITE) {
		
		// allocate new block for indirect3
		int newBlkNum  = getNewBlock(); 
		if (newBlkNum < 0 ){ // when no free block avaliable error !
		    System.out.println("getBlockNum Error: Disk is Full, no new free block. Operation: "+ operation);
		    return -1;
		}
		else {
		    inode.indirect3 = (short) newBlkNum; 
		    storeInode(inode);
		}
	    }
	    
	    // case 3
	    if (inode.indirect3 == 0 && operation == READ)
		// tell the caller it is case 3 by return zero (the pointer of indirect2)
		// the caller will handle the case (3)
		return inode.indirect3;
	    else { // case 1 and 2
		
		// get the lvl1 reference indirect block 
		// note: indirect3 is the top lvl (lvl0)
		int lvl1ref = dereference(operation, inode.indirect2, toplvlOffset);
		
		// If this block is unallocated, and operation == WRITE,
		// it will be allocated. 
		if (lvl1ref == 0 && operation == READ) {
		    return 0;
		}

		// get the lvl2 reference indirect block
		int lvl2ref = dereference(operation, lvl1ref, lvl1Offset);
		
		if (lvl2ref == 0 && operation == READ) {
		    return 0;
		}		

		// find or add the data block to the appropriate position
		return dereference(operation, lvl2ref, lvl2Offset);
	    }
	    
	}
	// ** DEBUG ** DEBUG ** DEBUG ** DEBUG ** DEBUG ** DEBUG **
	// Java reported a missing case specification here.
	// 
	System.out.println("Warning: getBlockNum() reached "+
			   "a case that is not handled");
	return -1;
    }
    
    

    // Returns a inumber (Inode number / file number) that is newly
    // allocated. The initial file size is zero.
    public int create() {
	Inode inode = null;
	for (int inum=0; inum < maxFiles; inum++) {
	    inode = namei(inum, READ);
	    if (inode == null) {
		inode = namei(inum, WRITE);
		if (inode != null) {
		    return inode.inumber;
		}
	    } 
	}
	return -1;
    }

    public int write(int inum, int offset, byte[] userbuf) {
	// Validate operands
	if (inum < 0 || inum >= maxFiles ||
	    offset < 0 ||
	    userbuf == null) {
	    System.out.println("Error write(arbitary range) : "+
			       "invalid operands.");
	    return -1;
	}
	if (((offset + userbuf.length) & (1 << 31)) != 0) {
	    System.out.println("Error write(arbitary range) : "+
			       "file cannot grow beyond 2 Gigabytes");
	}
	// Locate file inode.
	//   the READ flag tells namei NOT to create a file if 
	//   the file does not exist.
	Inode inode = namei(inum, READ);
	if (inode == null) {
	    // no such file: return with error.
	    System.out.println("Error write(arbitary range) : "+
			       "file (inumber="+inum+") does not exist");
	    return -1;
	}

	// Actually "writing zero-bytes" is a special case but
	// it is (1) sensible and (2) useful for creating an empty file
	// with an initial size.
	if (userbuf.length == 0) {
	    if (offset > inode.length) {
		inode.length = offset;
	    }
	    System.out.println("write(arbitary range) : "+
			       "writing zero bytes to file; "+
			       "new file size = "+inode.length+" bytes");
	    storeInode(inode);
	    return 0;
	}
	
	// assert(userbuf.length != 0);
	//
	int endOffset = offset + userbuf.length - 1;
	int startBlock = offset / blockSize;
	int startByte = offset % blockSize;
	int endBlock = endOffset / blockSize;
	int endByte = endOffset % blockSize;


	// Grow file size if necessary.
	if (endOffset >= inode.length) {
	    inode.length = endOffset + 1;
	    storeInode(inode);
	}

	// Write user's data into file.
	byte[] diskbuf = new byte[blockSize];

	int nextUserbufOffset = 0;
	// For eah block, do write. (vbn = Virtual Block Number)
	for (int vbn = startBlock; vbn <= endBlock; vbn++) {
	    // The number of bytes to write in this block
	    int thisWriteSize = -1;
	    // pbn = Physical Block Number
	    int pbn = getBlockNum(WRITE, inode, vbn);
	    if (pbn <= 0) {
		System.out.println("Error write(arbitary range): "+
				   "cannot allocate new block "+
				   "(possibly disk full).");
		// abruptly exit with the number of bytes already written
		return nextUserbufOffset; 
	    }
	    // First block is a special case.
	    if (vbn == startBlock) {
		if (userbuf.length > blockSize) 
		    thisWriteSize = blockSize;
		else
		    thisWriteSize = userbuf.length;
		if (startByte + userbuf.length > blockSize)
		    thisWriteSize = blockSize - startByte; 
		disk.read(pbn, diskbuf);
		System.arraycopy(userbuf, 0,
				 diskbuf, startByte,
				 thisWriteSize); 
		disk.write(pbn, diskbuf);
		System.out.println("Writing "+thisWriteSize+
				   " bytes to (vbn="+vbn+", pbn="+
				   pbn+", byte offset="+startByte+")");
		nextUserbufOffset = thisWriteSize;
	    } 
	    // Ordinary blocks
	    else if (vbn != endBlock) {
		thisWriteSize = blockSize;
		System.arraycopy(userbuf, nextUserbufOffset,
				 diskbuf, 0,
				 thisWriteSize);
		disk.write(pbn, diskbuf);
		System.out.println("Writing "+thisWriteSize+
				   " bytes to (vbn="+vbn+", pbn="+
				   pbn+", byte offset="+0+")");
		nextUserbufOffset += thisWriteSize;
	    }
	    // Last block is also a special case.
	    else /* (vbn == endBlock) */ {
		thisWriteSize = userbuf.length - nextUserbufOffset;
		disk.read(pbn, diskbuf);
		System.arraycopy(userbuf, nextUserbufOffset,
				 diskbuf, 0,
				 thisWriteSize);
		disk.write(pbn, diskbuf);
		System.out.println("Writing "+thisWriteSize+
				   " bytes to (vbn="+vbn+", pbn="+
				   pbn+", byte offset="+0+")");
		nextUserbufOffset += thisWriteSize;
	    }
	}
	
	// return the number of bytes actually written.
	return nextUserbufOffset;
    }

    /*
     *
     *
     *
     *
     *
     *
     */
    public int read(int inum, int offset, byte[] userbuf) {
	// Validate operands
	if (inum < 0 || inum >= maxFiles ||
	    offset < 0 ||
	    userbuf == null) {
	    System.out.println("Error read(arbitary range) : "+
			       "invalid operands.");
	    return -1;
	}
	// Locate file inode.
	//   the READ flag tells namei NOT to create a file if 
	//   the file does not exist.
	Inode inode = namei(inum, READ);
	if (inode == null) {
	    // no such file: return with error.
	    System.out.println("Error read(arbitary range) : "+
			       "file (inumber="+inum+") does not exist");
	    return -1;
	}

	// "Reading zero bytes" does not make sense. 
	if (userbuf.length == 0) {
	    System.out.println("read(arbitary range) : "+
			       "reading zero bytes from file.");
	    return 0;
	}
	
	// assert(userbuf.length != 0);
	//

	// check if the starting point is outside the file size.
	if (offset >= inode.length) {
	    return 0;
	}

	int endOffset = offset + userbuf.length - 1;
	
	// check if the end is outside the file size.
	// The check (endOffset < 0) actually targets addresses larger
	// than 2GB (which will appear as negative numbers due to overflow)
	// and trims down the ending offset if necessary.
	if (endOffset >= inode.length || endOffset < 0) {
	    endOffset = inode.length - 1;
	}

	// compute other parameters.
	int startBlock = offset / blockSize;
	int startByte = offset % blockSize;
	int endBlock = endOffset / blockSize;
	int endByte = endOffset % blockSize;

	// Read user's data from file.
	byte[] diskbuf = new byte[blockSize];

	int nextUserbufOffset = 0;
	// For eah block, do write. (vbn = Virtual Block Number)
	for (int vbn = startBlock; vbn <= endBlock; vbn++) {
	    // The number of bytes to write in this block
	    int thisReadSize = -1;
	    // pbn = Physical Block Number
	    int pbn = getBlockNum(READ, inode, vbn);
	    // First block is a special case.
	    if (vbn == startBlock) {
		if (vbn != endBlock) 
		    thisReadSize = blockSize;
		else
		    thisReadSize = endByte - startByte + 1; 
		if (pbn > 0) {
		    disk.read(pbn, diskbuf);
		} else {
		    fillByte(diskbuf, 0);
		}
		System.arraycopy(diskbuf, startByte,
				 userbuf, 0,
				 thisReadSize); 
		System.out.println("Reading "+thisReadSize+
				   " bytes from (vbn="+vbn+", pbn="+
				   pbn+", byte offset="+startByte+")");
		nextUserbufOffset = thisReadSize;
	    } 
	    // Ordinary blocks
	    else if (vbn != endBlock) {
		thisReadSize = blockSize;
		if (pbn > 0) {
		    disk.read(pbn, diskbuf);
		} else {
		    fillByte(diskbuf, 0);
		}
		System.arraycopy(diskbuf, 0,
				 userbuf, nextUserbufOffset,
				 thisReadSize);
		System.out.println("Reading "+thisReadSize+
				   " bytes from (vbn="+vbn+", pbn="+
				   pbn+", byte offset="+0+")");
		nextUserbufOffset += thisReadSize;
	    }
	    // Last block is also a special case.
	    else /* (vbn == endBlock) */ {
		thisReadSize =  endByte + 1;
		if (pbn > 0) {
		    disk.read(pbn, diskbuf);
		} else {
		    fillByte(diskbuf, 0);
		}
		System.arraycopy(diskbuf, 0,
				 userbuf, nextUserbufOffset,
				 thisReadSize);
		System.out.println("Reading "+thisReadSize+
				   " bytes from (vbn="+vbn+", pbn="+
				   pbn+", byte offset="+0+")");
		nextUserbufOffset += thisReadSize;
	    }
	}
	
	// return the number of bytes actually written.
	return nextUserbufOffset;
    }

    /***
     * delete the file indicate by the filenumber
     * Find the inode correspond to the file, deallocate all the data and
     * indirect blocks point by the inode by marking "free" (0) in the free
     * map. Moreover, the inode is marked "unused" by setting length = -1
     *
     * NOTE: The inode block or indirect block pointing to the inode is not
     * deallocated.
     *
     * @param fileNumber the inumber of the file being delete
     * @return int -1 for the file being delete is not exist
     *
     */
    public int delete(int fileNumber) {
	if (fileNumber < 0 || fileNumber >= maxFiles) {
	    System.out.println("delete error: invalid fileNumber ("+
			       fileNumber +")");
	    return -1;
	}
	
	// locate the inode
	Inode inode = namei(fileNumber, READ);
	if (inode == null) { // file does not exist
	    return -1;
	}
	
	// access each data block of the file to free it
	int numBlk = (inode.length / blockSize) +
	    ( ((inode.length % blockSize) == 0) ? (0) : (1) );
	System.out.println("delete(int) : deleting file "+fileNumber+
			   ", virtual block number from 0 to "+numBlk);
	for (int i=0; i<=numBlk; i++) {
	    int blkNum = getBlockNum(READ, inode, i);
	    System.out.println("delete(int) : (vbn "+i+" = pbn "+blkNum+")");
	    if (blkNum == -1) {
		System.out.println("delete: getBlockNum: error");
		return -1;
	    }
	    else {
		// check if the block is allocated, if yes, free it
		checkNFreeBlk(blkNum);
	    }
	}
	
	// free the indirect blocks inside

	// deallocate indirect1
	checkNFreeBlk(inode.indirect1); // single lvl

	// deallocate all the block pointing from indirect2
	if (inode.indirect2 !=0) { // 2 lvl
	    // iterate through all the pointer in indirect2
	    for (int i=0; i<ptrPerBlock; i++){
		int tmpBlkNum = dereference(READ, inode.indirect2, i);
		checkNFreeBlk(tmpBlkNum);
	    }
	    // deallcated indirect2 itself
	    checkNFreeBlk(inode.indirect2);
	}
	
	// deallcate all the block pointing from indirect3
	if (inode.indirect3 !=0) { // 3 lvl
	    for (int i=0; i<ptrPerBlock;i++){
		int tmpBlkNum2 = dereference(READ, inode.indirect3, i);
		for (int j=0; i<ptrPerBlock; j++){
		    int tmpBlkNum1 = dereference(READ, tmpBlkNum2, j);
		    checkNFreeBlk(tmpBlkNum1);
		}
		checkNFreeBlk(tmpBlkNum2);
	    }
	    checkNFreeBlk(inode.indirect3);
	}

	// mark inode as "unused"
	inode.length = -1;
	
	// update the state of this inode in the InodeBlock
	byte[] buffer = new byte[blockSize];
	
	disk.read(inode.inodeBlock, buffer);
	// Save this inode record into the Inode-block.
	inode.pack(buffer, inode.inodeBlockOffset);
	disk.write(inode.inodeBlock, buffer);
	System.out.println("block["+inode.inodeBlock+"].inode["+
			   inode.inodeBlockOffset+"]= (inumber="+
			   inode.inumber+"; size="+
			   inode.length+")");	
	
	// finish delete , return 
	return 0;
    }

    public void checkNFreeBlk(int pointer){
	if (pointer == -1){
	    System.out.println("Some Error Occur in checkNFreeBlk(" 
			       + pointer+ ")" );
	}
	if (pointer != 0){
	    int checkError = releaseBlock(pointer);
	    if (checkError == -1) {
		System.out.println("checkNFreeBlk: Cannot release block ("+
				   pointer +")");
	    } else {
		System.out.println("checkNFreeBlk: freed block "+pointer);
	    }
	}
    }

    public int size(int fileNumber){
	if (fileNumber < 0 || fileNumber >= maxFiles) {
	    System.out.println("delete error: invalid fileNumber ("+
			       fileNumber +")");
	    return -1;
	}
	
	// locate the inode
	Inode inode = namei(fileNumber, READ);
	if (inode == null) { // file does not exist
	    return -1;
	}

	return inode.length;
    }

} // end of FileSystem.java









