import java.io.*;
import java.util.*;

/** Basic driver program to be used as a shell for the MiniKernel for project 5.
 * It can be run in two modes:
 * <dl compact>
 *        <dt>Interactive:              <dd>java Boot ... FileTester
 *        <dt>With a test script file:  <dd>java Boot ... FileTester script
 * </dl>
 * To get a list of supported commands, type 'help' at the command prompt.
 * <p>
 * The testfile consists of commands to the driver program (one per line) as
 * well as comments.  Comments beginning with /* will be ignored completely by
 * the driver.  Comments beginning with // will be echoed to the output.
 * A command may optionally be preceded by "var=", in which case, the result
 * will be assigned to shell variable var.  Shell variables can be accessed
 * with the notation "$var".
 * <p>
 * See the test files test*.data for examples.
 */
public class FileTester {
    /** Synopsis of commands. */
    private static String[] helpInfo = {
        "help",
        "quit",
        "vars",
        "format maxFiles",
        "read inum offset size",
        "write inum offset size pattern ...",
        "writeln inum offset",
        "create",
        "delete inum",
        "size inum"
    };

    /** Values of variables (String to String). */
    private static Map vars = new HashMap();

    /** Main program.
     * @param args command-line arguments (there should be at most one:
     *      the name of a test file from which to read commands).
     */
    public static void main(String [] args){
        // NB:  This program is designed only to test the file system support
        // of the kernel, so it "cheats" in using non-kernel operations to
        // read commands and write diagnostics.
        if (args.length > 1) {
            System.err.println("usage: FileTester [ script-file ]");
            System.exit(0);
        }

        // Is the input coming from a file?
        boolean fromFile = (args.length == 1);

        // Create a stream for input
        BufferedReader input = null;

        // Open our input stream
        if (fromFile) {
            try {
                input = new BufferedReader(new FileReader(args[0]));
            } catch (FileNotFoundException e) {
                System.err.println("Error: Script file "
                        + args[0] + " not found.");
                System.exit(1);
            }
        } else {
            input = new BufferedReader(new InputStreamReader(System.in));
        }

        // Cycle through user or file input
        for (;;) {
            String cmd = null;
            try {
                // Print out the prompt for the user
                if (!fromFile) {
                    System.out.print("--> ");
                    System.out.flush();
                }

                // Read in a line
                String line = input.readLine();

                // Check for EOF and empty lines
                if (line == null) {
                    // End of file (Ctrl-D for interactive input)
                    return;
                }
                line = line.trim();
                if (line.length() == 0) {
                    pl("");
                    continue;
                }

                // Handle comments and echoing
                if (line.startsWith("//")) {
                    if (fromFile) {
                        pl(line);
                    }
                    continue;
                }
                if (line.startsWith("/*")) {
                    continue;
                }

                // Echo the command line
                if (fromFile) {
                    pl("--> " + line);
                }

                // Parse the command line
                StringBuffer sb = new StringBuffer(line);

                // See if the command has the form "var = ..."
                String dest = null;
                int pos = sb.indexOf("=");
                if (pos >= 0) {
                    dest = sb.substring(0, pos).trim();
                    if (dest.length() < 1) {
                        pl("bad assignment syntax");
                        continue;
                    }
                    sb.delete(0, pos + 1);
                }
                cmd = nextToken(sb);
                if (!expandVars(sb)) {
                    continue;
                }

                // Call the function that corresponds to the command
                int result = 0;
                if (cmd.equalsIgnoreCase("quit")) {
                    return;
                } else if (cmd.equalsIgnoreCase("help") || cmd.equals("?")) {
                    help();
                    continue;
                } else if (cmd.equalsIgnoreCase("vars")) {
                    for (Iterator i = vars.keySet().iterator(); i.hasNext(); ) {
                        Object o = i.next();
                        pl(o + " = " + vars.get(o));
                    }
                    continue;
                } else if (cmd.equalsIgnoreCase("format")) {
                    result = Library.format(nextInt(sb));
                } else if (cmd.equalsIgnoreCase("read")) {
                    int inum = nextInt(sb);
                    int offset = nextInt(sb);
                    int size = nextInt(sb);
                    result = readTest(inum, offset, size);
                } else if (cmd.equalsIgnoreCase("write")) {
                    int inum = nextInt(sb);
                    int offset = nextInt(sb);
                    int size = nextInt(sb);
                    result = writeTest(inum, offset, size, sb.toString());
                } else if (cmd.equalsIgnoreCase("writeln")) {
                    int inum = nextInt(sb);
                    int offset = nextInt(sb);
                    result = writeLines(inum, offset, input);
                } else if (cmd.equalsIgnoreCase("create")) {
                    result = Library.create();
                } else if (cmd.equalsIgnoreCase("delete")) {
                    result = Library.delete(nextInt(sb));
                } else if (cmd.equalsIgnoreCase("size")) {
                    result = Library.size(nextInt(sb));
                } else {
                    pl("unknown command");
                    continue;
                }

                // Print out the result of the function call
                pl("    Result is " + result);
                if (dest != null) {
                    vars.put(dest, Integer.toString(result));
                }
            } catch (NumberFormatException e) {
                // Handler for Integer.parseInt(...)
                pl("Incorrect argument type");
                help(cmd);
            } catch (NoSuchElementException e) {
                // Handler for nextToken()
                pl("Incorrect number of arguments");
                help(cmd);
            } catch (IOException e) {
                e.printStackTrace();
                return;
            }
        } // for (;;)
    } // main(String[])

    /** Helper function for main, to interpret a command argument.
     * Removes the next (space-separated) argument from the command line
     * and returns it, with surrounding spaces removed.
     * @param sb the remaining command line, updated in place.
     * @return the next token.
     * @throws NoSuchElementException is there are no more tokens.
     */
    private static String nextToken(StringBuffer sb) {
        int start = 0;
        int len = sb.length();
        while (start < len && Character.isWhitespace(sb.charAt(start))) {
            start++;
        }
        if (start >= len) {
            throw new NoSuchElementException();
        }
        int end = start + 1;
        while (end < len && !Character.isWhitespace(sb.charAt(end))) {
            end++;
        }
        String result = sb.substring(start, end);
        while (end < len && Character.isWhitespace(sb.charAt(end))) {
            end++;
        }
        sb.delete(0, end);
        return result;
    } // nextToken(StringBuffer)

    /** Helper function for main, to interpret a command argument.
     * Removes the next (space-separated) argument from the command line
     * and returns it its integer value.
     * @param sb the remaining command line, updated in place.
     * @return the value of the next token.
     * @throws NoSuchElementException is there are no more tokens.
     * @throws NumberFormatException if the next token is not numeric.
     */
    private static int nextInt(StringBuffer sb) {
        return Integer.parseInt(nextToken(sb));
    } // nextInt(StringBuffer)

    /** Helper function for main.  Expand variable references of the form
     * "$var".
     * @param sb the command line to be updated (in place).
     * @return false if there is some sort of error (also print a message).
     */
    private static boolean expandVars(StringBuffer sb) {
        for (;;) {
            int len = sb.length();
            int pos = sb.indexOf("$");
            if (pos < 0) {
                return true;
            }
            if (pos >= len - 1) {
                pl("bad variable syntax");
                return false;
            }
            int end = pos + 2;
            while (end < len && Character.isJavaIdentifierPart(sb.charAt(end)))
            {
                end++;
            }
            String var = sb.substring(pos + 1, end);
            if (!vars.containsKey(var)) {
                pl(var + ": undefined variable");
                return false;
            }
            sb.replace(pos, end, (String) vars.get(var));
        }
    } // expandVars(StringBuffer)

    /** Prints a list of available commands. */
    private static void help() {
        for (int i = 0; i < helpInfo.length; i++) {
            pl("    " + helpInfo[i]);
        }
    } // help()

    /** Prints help for command "cmd".
     * @param cmd the name of the command.
     */
    private static void help(String cmd) {
        for (int i = 0; i < helpInfo.length; i++) {
            if (helpInfo[i].startsWith(cmd)) {
                pl("usage: " + helpInfo[i]);
                return;
            }
        }
        pl("unknown command '" + cmd + "'");
    } // help(String)

    /** Reads data from a (simulated) file using Library.read
     * and displays the results.
     * @param inum the inumber of the file.
     * @param offset the offset from the start of the file, in bytes.
     * @param size amount of data to read, in bytes.
     * @return the result of the Library.read call.
     */
    private static int readTest(int inum, int offset, int size) {
        byte[] buffer = new byte[size];
        int bytesRead;

        // Fill buffer with junk first
        Arrays.fill(buffer, (byte) '*');
        bytesRead = Library.read(inum, offset, buffer);
        for (int i = 0; i < bytesRead; i++) {
            showChar(buffer[i] & 0xff);
        }
        pl("");
        return bytesRead;
    } // readTest(int, int, int)

    /** Writes data to a (simulated) file using Library.write.
     * @param inum the inumber of output file.
     * @param offset starting offset from the beginning of the file, in bytes.
     * @param size the number of bytes to write.
     * @param pattern source of data to write.
     * @return the result of the Library.write call.
     */
    private static int writeTest(
            int inum, int offset, int size, String pattern)
    {
        byte[] buffer = new byte[size];
        for (int i = 0; i < buffer.length; i++) {
            buffer[i] = (byte) pattern.charAt(i % pattern.length());
        }

        return Library.write(inum, offset, buffer);
    } // writeTest(int, int, int, String)

    /** Write data to a (simulated) file using Library.write.
     * Data comes from the following lines in the input stream.
     * @param inum the inumber of the output file.
     * @param offset starting offset from the beginning of the file, in bytes.
     * @param input source of the data to write.
     * @return the result of the Library.write call.
     */
    private static int writeLines(
            int inum, int offset, BufferedReader input)
            throws IOException
    {
        StringBuffer sb = new StringBuffer();
        for (;;) {
            String line = input.readLine();
            if (line == null || line.equals(".")) {
                break;
            }
            sb.append(line);
            sb.append("\n");
        }
        byte[] buffer = new byte[sb.length()];
        for (int i = 0; i < buffer.length; i++) {
            buffer[i] = (byte) sb.charAt(i);
        }
        return Library.write(inum, offset, buffer);
    } // writeLines(int, int, BufferedReader)

    /** Display a readable representation of a byte.
     * @param b the byte to display as a number in the range 0..255.
     */
    private static void showChar(int b) {
        if (b >= ' ' && b <= '~') {
            System.out.print((char)b);
            return;
        }
        if (b == '\n') {
            System.out.print('\n');
            return;
        }
        if (b == '\\') {
            System.out.print("\\\\");
            return;
        }
        System.out.print('\\');
        System.out.print(Integer.toString(b, 8));
    } // showChar(int)

    /** Print a line to System.out.
     * @param o the message to print.
     */
    private static void pl(Object o) {
        System.out.println(o);
    } // pl(Object)
} // FileTester
