/** In-memory representation of an inode */
// an inode is a file
public class Inode {
    /* Size of this file, in bytes (-1 means unused) */
    public int length;
    
    /* The inumber of the file store in this inode */
    public int inumber;

    /* The InodeBlock pointer to this Inode */
    public int inodeBlock;

    /* The Offset in the InodeBlock correspond to the Inode */
    public int inodeBlockOffset;

    /* Pointers to the first 11 data blocks */
    public short direct[] = new short[11];

    /* Pointer to the an indirect block */
    public short indirect1;

    /* Pointer to a doubly indirect block */
    public short indirect2;

    /* Pointer to a triply indirect block */
    public short indirect3;

    // constructor
    // call when "creat" file
    // all pointer are initialized to null
    // length (filesize) is zero
    // Note: The user of the Inode is responsible for updating the pointer
    // and the length when accessing the Inode (file)
    public Inode(){ 
	length = 0;
	inumber = 0;
	for (int i = 0; i < 11; i++){
	    direct[i] = 0; 
	}
	indirect1 = 0;
	indirect2 = 0;
	indirect3 = 0;
    }

    /** Pack this Inode into a byte array for writing to disk.
     * @param buf    an array to be written to disk.
     * @param offset the offset in buf, in units of inodes (the first inode
     *               goes at offset 0, the second at offset 1, etc.)
     */
    public void pack(byte[] buf, int offset) {
	// putting the inode information to the offset inode of the byte
	// array
	// e.g  buf[some offset] = length ..etc..
	// to update the block in the DISK
	
	int refOffset = offset*32; 
	
	// pack "length" (32-bit)
	buf[refOffset++] = (byte) (length >> 24);
	buf[refOffset++] = (byte) (length >> 16);
	buf[refOffset++] = (byte) (length >> 8);
	buf[refOffset++] = (byte) length;
	
	// pack "direct[]" (each 16-bit)
	for (int i=0; i<11;i++){
	    buf[refOffset++] = (byte) (direct[i] >> 8);
	    buf[refOffset++] = (byte) direct[i];
	}
	
	// pack "indirect1,2,3" (each 16-bit)
	buf[refOffset++] = (byte) (indirect1 >> 8);
	buf[refOffset++] = (byte) indirect1;
	buf[refOffset++] = (byte) (indirect2 >> 8);
	buf[refOffset++] = (byte) indirect2;
	buf[refOffset++] = (byte) (indirect3 >> 8);
	buf[refOffset++] = (byte) indirect3;

    }

    /** Initialize this Inode from information read from disk.
     * @param buf    an in-memory copy of a disk block.
     * @param offset the offset in buf, in units of inodes (the first inode
     *               goes at offset 0, the second at offset 1, etc.)
     */
    public void unpack(byte[] buf, int offset) {
	// putting back the file information from the offset inode in 
	// in the inode block
	// e.g  length = buf[some offset]
	
	int refOffset = offset*32;
	
	// unpack "length"
	length = (int) ((buf[refOffset++] & 0xff) << 24) + 
	    ((buf[refOffset++] & 0xff) << 16) + 
	    ((buf[refOffset++] & 0xff) << 8) +
	    (buf[refOffset++] & 0xff);
	
	// unpack "direct[]"
	for (int i=0; i<11; i++){
	    direct[i] = (short)( ((buf[refOffset++] & 0xff) << 8) + (buf[refOffset++] & 0xff) );
	}

	// unpack "indirect1,2,3"
	indirect1 = (short)( ((buf[refOffset++] & 0xff) << 8) + (buf[refOffset++] & 0xff) );
	indirect2 = (short)( ((buf[refOffset++] & 0xff) << 8) + (buf[refOffset++] & 0xff) );
	indirect3 = (short)( ((buf[refOffset++] & 0xff) << 8) + (buf[refOffset++] & 0xff) );
    }
}

