/* $Id: Library.p5.java,v 1.1 2002/11/24 12:12:42 solomon Exp solomon $ */

/** Convenience calls for using the Kernel.
 * Each function in this class makes a system call.  Sometimes, the arguments
 * are manipulated to make their user representation more convenient.
 * Note that this class contains only static methods.
 * All methods return integers.  Negative return values are error codes.
 * Some methods return positive values; others simply return 0 to mean "ok".
 *
 * @see Kernel
 */
public class Library {
    /** This private constructor ensures that no instances of Library are
     * ever created.
     */
    private Library() {}

    /** A table of error messages corresponding to Kernel error return codes.
     * This table should be indexed by the negative of rc, where
     * <pre>
     *          rc = Kernel.interrupt(Kernel.INTERRUPT_USER, ... )
     * </pre>
     * and rc is less than 0.
     */
    public static final String[] errorMessage = {
        "OK", // 0
        "Invalid argument type", // ERROR_BAD_ARGUMENT_TYPE = -1
        "No such class", // ERROR_BAD_COMMAND = -2
        "Argument out of range", // ERROR_OUT_OF_RANGE = -3
        "End of file on console input", // ERROR_END_OF_FILE = -4
        "I/O error on console input", // ERROR_IO = -5
        "Exception in user program", // ERROR_IN_CHILD = -6
        "No such process" // ERROR_NO_SUCH_PROCESS = -7
    };

    /** Performs SYSCALL_OUTPUT.
     * Displays text on the console.
     * @param s a String to display
     * @return zero
     */
    public static int output(String s) {
        return Kernel.interrupt(Kernel.INTERRUPT_USER,
            Kernel.SYSCALL_OUTPUT, 0, s, null, null);
    } // output

    /** Performs SYSCALL_INPUT.
     * Waits for the user to type some text and hit [return].
     * The input line is returned in the supplied StringBuffer
     * @param result a place to put the result
     * @return zero on success, or one of the error codes Kernel.END_OF_FILE or
     * Kernel.ERROR_IO.
     */
    public static int input(StringBuffer result) {
        result.setLength(0);
        return Kernel.interrupt(Kernel.INTERRUPT_USER,
                            Kernel.SYSCALL_INPUT, 0, result, null, null);
    } // input

    /** Performs SYSCALL_EXEC.
     * Launches the named program, and lets it run in parallel
     * to the current program.
     * @param command The name of a Java class to execute.
     * @param args The arguments to give the new program
     * @return a non-negative process id, or ERROR_BAD_COMMAND.
     */
    public static int exec(String command, String args[]) {
        return Kernel.interrupt(Kernel.INTERRUPT_USER,
            Kernel.SYSCALL_EXEC, 0, command, args, null);
    } // exec

    /** Performs SYSCALL_JOIN
     * Waits for a process to terminate
     * @param pid a process id returned by a previous call to exec.
     * @return Zero or ERROR_NO_SUCH_PROCESS
     */
    public static int join(int pid) {
        return Kernel.interrupt(Kernel.INTERRUPT_USER,
            Kernel.SYSCALL_JOIN, pid, null, null, null);
    } // exec

    /** Format the disk.  If the disk is already formatted, this system call
     * will destroy all data on it.
     * @param   maxFiles the maximum number of files the disk can hold.
     * @return  0 on success, -1 on failure.
     */
    public static int format(int maxFiles) {
        return Kernel.interrupt(Kernel.INTERRUPT_USER,
				Kernel.SYSCALL_FORMAT,
				maxFiles, null, null, null);
    } // format(int)

    /** Reads up to buffer.length bytes from a file.
     * @param inum      the inumber of the file to read from.
     * @param offset    the starting offset (in bytes from the beginning of the
     *                  file).
     * @param buffer    the destination for the data.
     * @return          the number of bytes read.  Will be at most
     *                  buffer.length, but may be smaller if offset is near the
     *                  end of the file.  Returns -1 for errors.
     */
    public static int read(int inum, int offset, byte[] buffer) {
	Integer inumObj=new Integer(inum);
	return Kernel.interrupt(Kernel.INTERRUPT_USER,
				Kernel.SYSCALL_READ,
				offset, inumObj, null, buffer);
    } // read(int, int, byte[])

    /** Writes up to buffer.length bytes to a file.  Existing data will be
     * overwritten.  The size of the file may be extended.
     * @param inum      the inumber of the file to write to.
     * @param offset    the starting offset (in bytes from the beginning of the
     *                  file).
     * @param buffer    the source of the data to write.
     * @return          the number of bytes written.  Will normally be
     *                  buffer.length.  Returns -1 for errors.
     */
    public static int write(int inum, int offset, byte[] buffer) {
	Integer inumObj=new Integer(inum);
	return Kernel.interrupt(Kernel.INTERRUPT_USER,
				Kernel.SYSCALL_WRITE,
				offset, inumObj, null, buffer);
    } // write(int, int, byte[])

    /** Creates a new file.
     * @return the inumber of a new file, or -1 on failure (too many files).
     */
    public static int create() {
	return Kernel.interrupt(Kernel.INTERRUPT_USER,
				Kernel.SYSCALL_CREATE,
				0, null, null, null);
    } // create()

    /** Deletes a file.
     * @param inum the inumber of the file to delete.
     * @return 0 on success, -1 on failure.
     */
    public static int delete(int inum) {
	Integer inumObj=new Integer(inum);
	return Kernel.interrupt(Kernel.INTERRUPT_USER,
				Kernel.SYSCALL_DELETE,
				0, inumObj, null, null);
    } // delete(int)

    /** Returns the size of a file.
     * @param inum the inumber of an existing file.
     * @return the currernt size of the file, in bytes.
     */
    public static int size(int inum) {
	Integer inumObj=new Integer(inum);
	return Kernel.interrupt(Kernel.INTERRUPT_USER,
				Kernel.SYSCALL_SIZE,
				0, inumObj, null, null);
    } // size(int)
} // Library
