/****************************************************************************
 * Block access within a file
 * takes the operation going to perform with the request block, 
 * an Inode and a block-offset of a block within the file 
 * (lets say block A)
 *  return a pointer (short) to block A 
 * (i.e. return the block number of block A in disk)
 * @param operation determine what to do if the require block do no exist
 * @param inode the inode reference to the request block
 * @param offset the block offset with in the file of the request block
 * return 0 if operation = READ and block not exist 
 *                      (caller will handle to output a zeroed buffer)
 * return -1 if there is an error occur
 *
 * assume offset within the possible reference range (handle by caller)
 *************************************************************************/

public short getBlockNum(int operation, Inode inode, int offset){
    if (operation != READ && operation != WRITE) {
	System.out.println("getBlockNum Error: Invalid Operation code (" +
			   operation +")");   
	return -1;
    } 

    if (offset < fileBlocksPerLevel[0]) { // lvl 0

	// check if the block exists
	if (inode.direct[offset] == 0 && operation == WRITE){ 
	    // block does not exists, allocate it in the disk
	    // not return error untill cannot allocate new block 
	    // (i.e. disk full)
	    
	    // allocate the new block
	    int newBlkNum = getNewBlock();
	    if (newBlkNum != -1){
		return (short) newBlkNum;
	    }
	    else{ // cannot allocate new free block (disk full), 
		// propagate the error
		return -1;
	    }
	}
	else {
	    // operation == READ or block exist
	    return inode.direct[offset];
	}
    }
    else /*lvl 1*/ if ( (offset >= fileBlocksPerLevel[0]) &&
			(offset < fileBlocksPerLevel[1]) ) { 

	/* 3 case here (1) indirect1 = 0 && WRITE
	               (2) indirect1 != 0
		       (3) indirect1 = 0 && READ
	*/

	int adjOffset = offset - fileBlocksPerLevel[0];
	
	// case 1
	if (inode.indirect1 == 0 && operation == WRITE) {

	    // allocate new block for indirect1
	    int newBlkNum  = (short) getNewBlock(); 
	    if (newBlkNum < 0 ){ // when no free block avaliable error !
		System.out.println("getBlockNum Error: Disk is Full, no new free block. Operation: "+ operation);
		return -1;
	    }
	    else { inode.indirect1 = newBlkNum; }
	}

	// case 3
	if (inode.indirect1 == 0 && operation == READ)
	    // tell the caller it is case 3 by return zero (the pointer of indirect1)
	    // the caller will handle the case (3)
	    return inode.indirect1;
	else { // case 1 and 2

	    // find data block (or get a free data block and reference to it in case 1)
	    return (short) dereference(operation, inode.indirect1, adjOffset);
	}
    }
    else /* lvl 2*/ if ( (offset >= fileBlocksPerLevel[1]) &&
			 (offset < fileBlocksPerLevel[2]) ) {
	 
	 /* 3 case here (1) indirect2 = 0 && WRITE
	                (2) indirect2 != 0
		        (3) indirect2 = 0 && READ
	 */
	
	int adjOffset = offset - fileBlocksPerLevel[1];
	
	// lvl1 Offset
	int lvl1Offset = adjOffset % ptrPerBlock;
	
	adjOffset /= ptrPerBlock;

	// top lvl (lvl0) Offset
	int toplvlOffset = adjOffset;

	
	// case 1
	if (inode.indirect2 == 0 && operation == WRITE) {
		
	    // allocate new block for indirect2
	    int newBlkNum  = (short) getNewBlock(); 
	    if (newBlkNum < 0 ){ // when no free block avaliable error !
		System.out.println("getBlockNum Error: Disk is Full, no new free block. Operation: "+ operation);
		return -1;
	    }
	    else { inode.indirect2 = newBlkNum; }
	}
	
	// case 3
	if (inode.indirect2 == 0 && operation == READ)
	    // tell the caller it is case 3 by return zero (the pointer of indirect2)
	    // the caller will handle the case (3)
	    return inode.indirect2;
	else { // case 1 and 2
	    
	    // get the lvl1 reference indirect block 
	    // note: indirect2 is the top lvl (lvl0)
	    int lvl1ref =  dereference(operation, inode.indirect2, toplvlOffset);
	    
	    // find or add data block in appropriate position
	    return (short) dereference(operation, lvl1ref, lvl1Offset);
	}
    }
    else /* lvl 3*/ if ( (offset >= fileBlocksPerLevel[2]) &&
			 (offset < fileBlocksPerLevel[3]) ) {
	 
	 /* 3 case here (1) indirect2 = 0 && WRITE
	                (2) indirect2 != 0
		        (3) indirect2 = 0 && READ
	 */

    int adjOffset = offset - fileBlocksPerLevel[1];
	
    // lvl2 Offset
    int lvl2Offset = adjOffset % ptrPerBlock;
	
    adjOffset /= ptrPerBlock;

    // lvl1 Offset
    int lvl1Offset = adjOffset % ptrPerBlock;
	
    adjOffset /= ptrPerBlock;
    
    // top lvl (lvl0) Offset
    int toplvlOffset = adjOffset;

    // case 1
	if (inode.indirect3 == 0 && operation == WRITE) {
		
	    // allocate new block for indirect3
	    int newBlkNum  = (short) getNewBlock(); 
	    if (newBlkNum < 0 ){ // when no free block avaliable error !
		System.out.println("getBlockNum Error: Disk is Full, no new free block. Operation: "+ operation);
		return -1;
	    }
	    else { inode.indirect3 = newBlkNum; }
	}
	
	// case 3
	if (inode.indirect3 == 0 && operation == READ)
	    // tell the caller it is case 3 by return zero (the pointer of indirect2)
	    // the caller will handle the case (3)
	    return inode.indirect3;
	else { // case 1 and 2
	    
	    // get the lvl1 reference indirect block 
	    // note: indirect3 is the top lvl (lvl0)
	    int lvl1ref   dereference(operation, inode.indirect2, toplvlOffset);
	    
	    // get the lvl2 reference indirect block
	    int lvl2ref = dereference(operation, lvl1ref, lvl1Offset);

	    // find or add the data block to the appropriate position
	    return (short) dereference(operation, lvl2ref, lvl2Offset);
	}

    }
}






